from jax.config import config
config.update("jax_enable_x64", True)
#from .magneticaxis import *
#from .fouriercurve import *
#from .objectives import *
from .config import *
#from .biotsavart import *
#from .curve import RotatedCurve
