PROTON_MASS = 1.67262192369e-27  # kg
NEUTRON_MASS = 1.67492749804e-27  # kg
ELEMENTARY_CHARGE = 1.602176634e-19  # C
ONE_EV = 1.602176634e-19  # J
ALPHA_PARTICLE_MASS = 2*PROTON_MASS + 2*NEUTRON_MASS
ALPHA_PARTICLE_CHARGE = 2*ELEMENTARY_CHARGE
FUSION_ALPHA_PARTICLE_ENERGY = 3.52e6 * ONE_EV
